import os
import json
from copy import deepcopy

from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
from sqlalchemy import Column, DateTime, String, Integer, func

Base = declarative_base()


class MPlugin(Base):
    __tablename__ =  'mplugin'
    plugin_name = Column(String(100), primary_key=True,)
    module_name = Column(String(1200))
    metadata_type = Column(String(120), default='table')
    # plugin_status = Column(String(200))
    plugin_spec_path = Column(String(200))
    notebook_template_path = Column(String(500))
    ds_param  = Column(String(4000)) #Customized parameter by each plugin, stored as a python dict object in json string.
    supported_ds_types = Column(String(200))
    sample_ds_url = Column(String(1000))
    author = Column(String(200))
    created_date = Column(DateTime, default=func.now())
    plugin_desc = Column(String(1000))
    def __repr__(self):
        return "<pluggin (name='%s', path='%s' )>" % (
            self.plugin_name, self.plugin_spec_path )


#Reflect each database table we need to use, using metadata
class MDatasource(Base):
    __tablename__ =  'mdatasource'
    #db_id = Column(Integer, primary_key=True, autoincrement=True)
    ds_name = Column(String(100), primary_key=True,)
    # db_trx_id = Column(Integer, default=-2) # This is to track current datasource id if it is sychonized in real time. For example, hdfs inotify txid
    nbr_of_tables = Column(Integer)
    nbr_of_columns = Column(Integer)
    ds_type = Column(String(100) )
    # ds_url = Column(String(1000))
    # table_group_name = Column(String(200), default='_NA')
    created_date = Column(DateTime, default=func.now())
    last_reflect_date = Column(DateTime)
    ds_param  = Column(String(4000)) #Customized parameter by each plugin, stored as a python dict object in json string.
    ds_desc = Column(String(1000))
    ds_tags = Column(String(2550))
    def __repr__(self):
        return "<database (name='%s', ds_type = %s,   last_reflect_date='%s')>" % (
            self.ds_name, self.ds_type,   self.last_reflect_date)


    @staticmethod
    def from_dict(ds_dict=None):
        ds =     MDatasource(# table_group_name=ds_dict['table_group_name'],
                             ds_name=ds_dict['ds_name'],
                             ds_type=ds_dict['ds_type'],
                             # ds_url=ds_dict['ds_url'],
                             ds_param=json.dumps(ds_dict['ds_param']),
                             nbr_of_tables=0,
                             nbr_of_columns=9,
                             ds_desc=ds_dict['ds_desc'] if 'ds_desc' in ds_dict.keys() else '',
                             created_date=func.now(),
                             )
        return ds


    def to_dict(self):
        ds_dict = deepcopy(self.__dict__)
        ds_dict['ds_param'] = json.loads(ds_dict['ds_param'])
        return ds_dict


class MTable(Base):
    __tablename__ = 'mtable'
    # table_id = Column(Integer, primary_key=True, autoincrement=True)
    # db_id = Column(Integer)
    ds_name = Column(String(100), primary_key=True,)
    table_group_name = Column(String(200), default='_NA')
    table_name = Column(String(255), primary_key=True)
    table_type = Column(String(255))
    table_comment = Column(String(5000))

#Reflect each database table we need to use, using metadata
class MColumn(Base):
    __tablename__ =  'mcolumn'
    # table_id = Column(Integer, primary_key=True)
    ds_name = Column(String(100), primary_key=True,)
    table_group_name = Column(String(200), default='_NA')
    table_name = Column(String(255), primary_key=True)
    column_name = Column(String(255), primary_key=True)
    column_seq = Column(Integer)
    column_type = Column(String(255))
    column_comment = Column(String(1000))
    column_tags = Column(String(2550))



plugin_name = 'sqlalchemy'
plugin_desc = 'Sqlalchemy reflect engine, which works for many mainstream databases, including Mysql, Sqlite, etc. For each db to work, you need to install pymysql/mysqlclient (python3) or mysql-python (python2) in advance.'
supported_ds_types = ['mysql','sqlite3']
required_packages = []


ds_param = {'schema_name': {'type':'String',
                             'full_name':'Schema Name to Reflect:',
                             'default_value': '_NA',
                             'desc':'''''',
                             },
            'connect_string': {'type':'String',
                             'full_name':'Database Connection String:',
                             'default_value': 'sqlite:////tmp/db1',
                             'desc':'''The sqlalchemy access URL.''',
                             },
            }


sample_ds_url = 'mysql://root:xyz@localhost/employees'
metadata_type='table'

author = 'duan'
class_name = 'sqlalchemyindex.SQLAlchemyReflectEngine'
notebook_template_path = 'show_sqlalchemy_table_template.ipynb'

from .sqlalchemyengine import SQLAlchemyReflectEngine as ReflectEngine



plugin_name = 'MS_SQL_Server'
plugin_desc = 'mssql can reflect Microsoft SQL Server engine based on sqlalchemyindex.'
supported_ds_types = ['oracle']
required_packages = ['pymssql']


ds_param = {'schema_name': {'type':'String',
                             'full_name':'Schema Name to Reflect:',
                            'default_value':'Not effective yet',
                             'desc':'Not effective yet',
                             },
            'connect_string': {'type':'String',
                             'full_name':'Database Connection String:',
                               'default_value': 'mssql+pymssql://sa:yourPassword@127.0.0.1/samd?charset=utf8',
                               'desc':'''The sqlalchemy access URL.''',
                             },
            }

metadata_type='table'

author = 'duan'
# class_name = 'sqlalchemyindex.SQLAlchemyReflectEngine'
notebook_template_path = 'show_sqlalchemy_table_template.ipynb'

from schemaindex.plugin.sqlalchemyindex import ReflectEngine as SqlAlchemyReflectEngine
# -----------------------------------
class ReflectEngine(SqlAlchemyReflectEngine):
    pass
#    def __init__(self,ds_dict=None, ds_name = None):
#        SqlAlchemyReflectEngine.__init__(ds_dict = ds_dict, ds_name = ds_name )



import os, sys
import traceback, logging

# Those are all options for end user to customize. Other internal configurations may be found in different classes.
cfg = {"web": {
                'port': 8088,
                'address': "localhost",
                'use_anonymous' : True
               },
        'database': {
                'ds_type': 'sqlite3',
                'sqlite_file':  os.path.join(os.path.abspath(os.path.join(os.path.dirname(__file__))), '..' , "schemaindex.sqlite3") ,
                'user': 'root',
                'passwd': 'my secret password',
                },
        "main": {"schemaflex_home": os.path.abspath(os.path.join(os.path.dirname(__file__), '..')),
                 "schemaflex_spec": os.path.abspath(os.path.join(os.path.dirname(__file__), '..','plugin')),
                 "schemaindex_notebooks": os.path.abspath(os.path.join(os.path.dirname(__file__), '..','notebooks')),
                 "schemaflex_text_index_path": os.path.abspath(os.path.join(os.path.dirname(__file__), '..','indexdir')),
                 'init_indicator_file': os.path.abspath(os.path.join(os.path.dirname(__file__), 'do_schemaindex_init')),
                 'interval_commit_trx_id': 60,
                 'search_result_limit': 100,
                 },
        "logging": {"log_file": "schemaindex.log",
                            "log_dir": "log",
                            "log_level": "DEBUG"  # DEBUG,INFO,WARNING,ERROR,CRITICAL
                            }
}

sys.path.append( cfg['main']['schemaflex_spec'])

if not os.path.exists(cfg['main']['schemaflex_spec']):
    os.mkdir(cfg['main']['schemaflex_spec'])


LOG_DIR = os.path.join(cfg['main']['schemaflex_home'],
                       cfg['logging']['log_dir'])
if not os.path.exists(LOG_DIR):
    os.mkdir(LOG_DIR)



current_logfile_path = os.path.join(LOG_DIR,
                                    cfg['logging']['log_file'])


class MyStreamHandler(logging.StreamHandler):
    def format(self, record):
        try:
            return logging.StreamHandler.format(self, record)
        except TypeError:
            # Print a stack trace that includes the original log call
            traceback.print_stack()

    def handleError(self, record):
        raise


logging.basicConfig(format='[%(asctime)s] {%(pathname)s:%(lineno)d}  - %(message)s',
                    level=logging.getLevelName(cfg['logging']['log_level']),
                    filename=current_logfile_path)

log = logging.getLogger('schemaflex_logger')
handler = MyStreamHandler()
log.addHandler(handler)

# logging.getLogger('schemaflex_logger').info('Logging initialization finished.')  # will not print anything


class SchemaIndexPluginError(Exception):
    pass





from __future__ import unicode_literals
from __future__ import print_function

#
import pytest

from schemaindex.app.config import cfg

from schemaindex.app.schemaindexapp import si_app
from schemaindex.app.pluginmanager import si_pm
import os

@pytest.fixture(scope='module')
def resource_a_setup(request):
    print('\nDoing setup by putting init file')
    db_file_path = cfg['database']['sqlite_file']

    # open(config['main']['init_indicator_file'], 'a').close()
    if os.path.exists(db_file_path):
        print('DB file is ready,  going to delete it .')
        os.remove(db_file_path)


    si_app.schemaindex_init(db_file_path = db_file_path)

    def fin():
        print ("\nDoing teardown")
    request.addfinalizer(fin)



ds_dict = {
    'ds_name': 'emp1',
    'ds_type': 'sqlalchemy',
    'ds_desc': 'created by unittest',
    'ds_param': {'connect_string': 'sqlite:////home/duan/schemaindex.sqlite3.goodbackup',
                 'schema_name': 'na',
                 }
}

def test_plugin_loading(resource_a_setup):
    list1 = si_app.get_plugin_name_list()
    print(list1)
    assert len(list1) == 0

    si_pm.scan_reflect_plugins()

    list1 = si_app.get_plugin_name_list()
    assert  'simplehdfsindex'  in list1



def test_plugin_mysql_add(resource_a_setup):
    dict1 = si_app.get_data_source_dict(ds_name=ds_dict['ds_name'])
    assert dict1 is None

    si_app.add_data_soruce(ds_dict)
    dict1 = si_app.get_data_source_dict(ds_name=ds_dict['ds_name'])
    assert dict1['ds_param']['connect_string'] == ds_dict['ds_param']['connect_string']

def test_plugin_mysql_reflect(after='test_plugin_mysql_add'):

    si_pm.reflect_db(data_source_name=ds_dict['ds_name'])
    suggest1=si_app.get_whoosh_search_suggestion(q='mta')
    print(suggest1)

    assert 'mtable' in suggest1



hdfs_ds_dict = {
    'ds_name': 'hdfs1',
    'ds_type': 'hdfsindex',
    'ds_desc': 'created by unittest of hdfsindex',
    'ds_url': 'http://localhost:50070',
    'ds_param': {'hdfs_web_url': 'http://localhost:50070',
                 'hdfs_url': 'hdfs://localhost:9000',
                 'start_inotify': 'off',
                 'root_path':'/',
                 'inotify_trx_id':'-1',
                 }
}


def test_plugin_hdfs_add():
    dict1 = si_app.get_data_source_dict(ds_name=hdfs_ds_dict['ds_name'])
    assert dict1 is None

    si_app.add_data_soruce(hdfs_ds_dict)
    dict1 = si_app.get_data_source_dict(ds_name=hdfs_ds_dict['ds_name'])
    assert dict1['ds_param']['hdfs_web_url'] ==  'http://localhost:50070'


def test_plugin_hdfs_reflect(after='test_plugin_hdfs_add_reflect'):
    si_pm.reflect_db(data_source_name=hdfs_ds_dict['ds_name'])
    suggest1 = si_app.get_whoosh_search_suggestion(q='citi')
    print(suggest1)
    assert 'cities' in suggest1
    #assert 'dept_name' in suggest1



import tornado.web
import json
from sqlalchemy.orm import create_session
from sqlalchemy import create_engine



from .dbmodels import MDatasource,MTable,MColumn

from .schemaindexapp import si_app
from .pluginmanager import si_pm
# si_app.datasource_init() # This call was in si_app.__init__, but it failed with "not able to load module". So i moved it here.


engine = create_engine('sqlite:///' + si_app.config['database']['sqlite_file'])


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.write("Hello, world")




class ReloadPluginHandler(tornado.web.RequestHandler):
    def get(self):
        si_pm.scan_reflect_plugins()
        session = create_session(bind=engine)
        dbrs = si_app.get_data_source_rs()
        ds_count = dbrs.count()

        base_navigation_dict = {'selected_menu': 'overview',
                                'dbrs': dbrs,
                                'plugin_list': si_app.get_plugin_list(),
                                'ds_count':str(ds_count),
                                #'tab_count':str(tab_count),
                                #'attr_count':str(attr_count),
                                #'tag_count':'NA',

        }

        self.render("overview.html", base_navigation_dict=base_navigation_dict)



class OverviewHandler(tornado.web.RequestHandler):
    def get(self):
        # self.write("Hello, world")
        # print('overview ed')
        session = create_session(bind=engine)
        dbrs = si_app.get_data_source_rs()

        ds_count = dbrs.count()
        #tab_count = session.query(MTable).count()
        #attr_count = session.query(MColumn).count()
        base_navigation_dict = {'selected_menu': 'overview',
                                'dbrs': dbrs,
                                'plugin_list': si_app.get_plugin_list(),
                                'ds_count':str(ds_count),
                                #'tab_count':str(tab_count),
                                #'attr_count':str(attr_count),
                                #'tag_count':'NA',

        }

        self.render("overview.html", base_navigation_dict=base_navigation_dict)


class GlobalSearchHandler(tornado.web.RequestHandler):
    def get(self):
        q = ''
        res = []
        dbrs = si_app.get_data_source_rs()

        base_navigation_dict = {'selected_menu': 'search',
                                'dbrs': dbrs,
                                'selected_schema_name': ''
                                }
        self.render("global_search_result.html",
                    dbrs=dbrs,
                    q=q,
                    base_navigation_dict=base_navigation_dict,
                    search_result = res)

    def post(self):

        q = self.get_argument('q')
        res = si_app.global_whoosh_search(q)
        dbrs = si_app.get_data_source_rs()

        base_navigation_dict = {'selected_menu': 'search',
                                'dbrs': dbrs,
                                'selected_schema_name': ''
                                }
        self.render("global_search_result.html",
                    dbrs=dbrs,
                    q=q,
                    base_navigation_dict=base_navigation_dict,
                    search_result = res)


class DatabaseSummaryHandler(tornado.web.RequestHandler):
    def get(self):
        # self.write("Hello, world")
        # param_db_name = self.get_argument("table_group_name", None)
        session = create_session(bind=engine)
        param_ds_name = self.get_argument("ds_name", None)
        if param_ds_name is None:
            si_app.logger.warn('error: no data source name is given (ds_name)')
            result_message =   {'message_type': 'danger',
                                'message_title': 'Error',
                                'message_body':'No data source name is given (ds_name)'}


            base_navigation_dict = {'selected_menu': 'datasources',
                                    'dbrs': session.query(MDatasource),
                                    'error': 'error: no schema name is given',
                                    'selected_add_data_source': False,
                                    'message': result_message,
                                    'db': None
                                    }
            # self.rend
            # er("404.html")
            self.render("database_summary.html",  # current_schema_name = param_schema_name,
                        base_navigation_dict=base_navigation_dict )

        entry = None

        dbrs1 = session.query(MDatasource).filter_by(ds_name = param_ds_name )
        db = dbrs1.first()
        if db is None:
            print('')
            si_app.logger.warn('error: did not find database' + param_ds_name)
            result_message =   {'message_type': 'danger',
                                'message_title': 'Error',
                                'message_body':'error: did not find database' + param_ds_name}


            base_navigation_dict = {'selected_menu': 'overview',
                                    'dbrs': session.query(MDatasource),
                                    'error': 'error: no schema name is given',
                                    'selected_add_data_source': False,
                                    'message': result_message,
                                    'plugin_list': si_app.get_plugin_list(),
                                    }
            # self.render("404.html")
            self.render("overview.html",  # current_schema_name = param_schema_name,
                        base_navigation_dict=base_navigation_dict )

        dbrs = si_app.get_data_source_rs()
        tabrs = si_app.get_table_list_for_data_source_rs(param_ds_name= param_ds_name)

        base_navigation_dict = {'selected_menu': 'datasources',
                                'selected_add_data_source': False,
                                'dbrs': dbrs,
                                'selected_schema_name': param_ds_name,
                                'db':db,
                                'tabrs':tabrs,
                                'plugin_name_list': si_app.get_plugin_name_list(),
                                }


        base_navigation_dict['input_db_type'] = db.ds_type
        plugin_info = si_app.get_plugin_info(p_plugin_name=db.ds_type )
        base_navigation_dict['input_ds_param'] = json.loads(plugin_info.ds_param)

        self.render("database_summary.html", # current_schema_name = param_schema_name,
                    base_navigation_dict=base_navigation_dict # ,dbrs=dbrs, tabrs=tabrs,db=db
                    )

    def post(self):
        ds_dict = {}
        ds_dict['ds_name'] = self.get_argument('ds_name')
        ds_dict['ds_type'] = self.get_argument('ds_type')
        ds_dict['ds_desc'] = self.get_argument('ds_desc')

        ds_dict['ds_param'] = {}
        ds_plugin_info = si_app.get_plugin_info(p_plugin_name=ds_dict['ds_type'])
        param_name_dict = json.loads(ds_plugin_info.ds_param)
        for param_name in param_name_dict.keys():
            if param_name_dict[param_name]['type'] == 'Boolean':
                ds_dict['ds_param'][param_name] = self.get_argument('ds_param.' + param_name, default='off' )
            else:
                ds_dict['ds_param'][param_name] = self.get_argument('ds_param.' + param_name)

        db = si_app.update_data_soruce(ds_dict)


        dbrs = si_app.get_data_source_rs()

        result_message = {'message_type': 'info',
                          'message_title': 'Success',
                          'message_body': 'The data source %s is updated.' % (
                          str(ds_dict['ds_name']) )}

        base_navigation_dict = {'selected_menu': 'database',
                                'dbrs': dbrs,
                                'selected_add_data_source':False,
                                'selected_schema_name':ds_dict['ds_name'],
                                'plugin_name_list': si_app.get_plugin_name_list(),
                                'message': result_message,
                                'db':db  # When 'db' is available, it is treated as show_data_source
                                }


        base_navigation_dict['input_db_type'] = db.ds_type
        plugin_info = si_app.get_plugin_info(p_plugin_name=db.ds_type )
        base_navigation_dict['input_ds_param'] = json.loads(plugin_info.ds_param)

        self.render("database_summary.html",
                    base_navigation_dict=base_navigation_dict, dbrs=dbrs, db = None,tabrs=None)


class AddDataSourceHandler(tornado.web.RequestHandler):
    def get(self):
        #
        # it is import not to put 'db' into this dict to indicate this is add_data_source
        base_navigation_dict = {'selected_menu': 'datasources',
                                'dbrs': si_app.get_data_source_rs(),
                                'plugin_name_list': si_app.get_plugin_name_list(),
                                'selected_add_data_source':True,
                                'input_ds_param':True,
                                'selected_schema_name':'__add_data_source__',
                                }
        ds_type = self.get_argument('ds_type',default=None)


        if ds_type is None:
            ds_type =si_app.get_plugin_name_list()[0]  # simply choose first available plugin


        base_navigation_dict['input_db_type'] = ds_type
        plugin_info = si_app.get_plugin_info(p_plugin_name=ds_type)
        base_navigation_dict['input_ds_param'] = json.loads(plugin_info.ds_param)

        self.render("database_summary.html",
                    base_navigation_dict=base_navigation_dict,
                      db=None,tabrs=None)


    def post(self):
        ds_dict = {}
        ds_dict['ds_name'] = self.get_argument('ds_name')
        ds_dict['ds_desc'] = self.get_argument('ds_desc')
        ds_dict['ds_type'] = self.get_argument('ds_type')

        ds_dict['reflect_database_automatic'] = self.get_argument('reflect_database_automatic', default='off')

        ds_dict['ds_param'] = {}
        ds_plugin_info = si_app.get_plugin_info(p_plugin_name=ds_dict['ds_type'])
        param_name_dict = json.loads(ds_plugin_info.ds_param)
        for param_name in param_name_dict.keys():
            if param_name_dict[param_name]['type'] == 'Boolean':
                ds_dict['ds_param'][param_name] = self.get_argument('ds_param.' + param_name, default=False )
            else:
                ds_dict['ds_param'][param_name] = self.get_argument('ds_param.' + param_name)

        try:
            si_app.add_data_soruce(ds_dict)
            if(ds_dict['reflect_database_automatic'] =='on'):
                si_pm.reflect_db(ds_dict['ds_name'])

            #Info = {'result': 'ok', 'message': 'A new data source is added.'}
            self.redirect('/database_summary?ds_name=' + ds_dict['ds_name'])
        except Exception as e:
            si_app.logger.error('failed to add data source!')
            si_app.logger.error(e)
            #Info = {'result': 'ok', 'message': 'A new data source is added.'}
            dbrs = si_app.get_data_source_rs()
            result_message =   {'message_type': 'danger',
                                'message_title': 'Error',
                                'message_body':'Failed to add data source: %s! The error was: %s' % (str(ds_dict['ds_name']), str(e))}

            base_navigation_dict = {'selected_menu': 'database',
                                    'dbrs': dbrs,
                                    'selected_add_data_source': True,
                                    'plugin_name_list': si_app.get_plugin_name_list(),
                                    'selected_schema_name': ds_dict['ds_name'],
                                    'message': result_message,
                                    'db': MDatasource.from_dict(ds_dict = ds_dict),
                                    }
            base_navigation_dict['input_db_type'] = ds_dict['ds_type']
            plugin_info = si_app.get_plugin_info(p_plugin_name=ds_dict['ds_type'])
            base_navigation_dict['input_ds_param'] = json.loads(plugin_info.ds_param)


            self.render("database_summary.html",
                        base_navigation_dict=base_navigation_dict)

        '''
        dbrs = si_app.get_data_source_rs()

        Info = {'result': 'ok', 'message':'A new data source is added.'}
        base_navigation_dict = {'selected_menu': 'database',
                                'dbrs': dbrs,
                                'plugin_name_list': si_app.get_plugin_name_list(),
                                'selected_schema_name':ds_dict['ds_name'],
                                'db':db,
                                }
        self.render("database_summary.html",
                    base_navigation_dict=base_navigation_dict)
        '''

class DeleteDataSourceHandler(tornado.web.RequestHandler):
    def post(self):
        ds_name = self.get_argument('ds_name')
        ds_dict  = si_app.get_data_source_dict(ds_name=ds_name)
        ds_dict['delete_reflected_database_automatic'] = self.get_argument('delete_reflected_database_automatic', default=None)

        result_message = si_app.delete_data_soruce(ds_dict)
        dbrs = si_app.get_data_source_rs()
        base_navigation_dict = {'selected_menu': 'database',
                                'dbrs': dbrs,
                                'selected_add_data_source':True,
                                'plugin_name_list': si_app.get_plugin_name_list(),
                                'selected_schema_name':ds_dict['ds_name'],
                                'message': result_message
                                }
        if result_message['message_title']  == 'Error':
            session = create_session(bind=engine)
            dbrs1 = session.query(MDatasource).filter_by(ds_name = ds_dict['ds_name'])
            db = dbrs1.first()
            base_navigation_dict['db'] = db
            base_navigation_dict['tabrs'] = si_app.get_table_list_for_data_source_rs(param_ds_name= ds_dict['ds_name'])
            base_navigation_dict['input_db_type'] = ds_dict['ds_type']
            plugin_info = si_app.get_plugin_info(p_plugin_name=ds_dict['ds_type'])
            base_navigation_dict['input_ds_param'] = json.loads(plugin_info.ds_param)

            self.render("database_summary.html",
                        base_navigation_dict=base_navigation_dict)

        self.redirect('/')




class ReflectDataSourceHandler(tornado.web.RequestHandler):

    def post(self):
        ds_name = self.get_argument('ds_name')
        si_pm.reflect_db(ds_name)
        # self.__setattr__('reflected_db', ds_name)
        self.redirect('/database_summary?ds_name='+ds_name)


class ViewTableInNotebookHandler(tornado.web.RequestHandler):
    def get(self):
        # json.dumps()
        table_id = self.get_argument('table_id')
        ds_name = self.get_argument('ds_name')


        gen_loc = si_pm.generate_notebook_for_table(table_id=table_id, ds_name=ds_name)
        print(gen_loc)
        from subprocess import Popen
        p = Popen(["jupyter", "notebook",  "--notebook-dir=" + si_app.config['main']['schemaflex_spec'], gen_loc]) # something long running

        Info = {'result': 'started in another browser'}
        self.write(json.dumps(Info))








class DatabaseJSONHandler(tornado.web.RequestHandler):
    def post(self):
        ds_dict = {}
        # ds_dict['table_group_name']  = self.get_argument('table_group_name')
        ds_dict['ds_name'] = self.get_argument('ds_name')
        ds_dict['ds_url'] = self.get_argument('ds_url')
        ds_dict['ds_type'] = self.get_argument('ds_type')

        si_app.add_data_soruce(ds_dict)


        Info = {'result': 'ok'}
        self.write(json.dumps(Info))

    def get(self):
        # json.dumps()
        print('si_app.datasource_init()')
        si_app.datasource_init()
        db_dict = {'ds_name':'emp1'}
        self.write(json.dumps(db_dict)) # data = json.dumps(db_dict)

class SearchSuggestionJSONHandler(tornado.web.RequestHandler):
    def get(self):
        q = self.get_argument('query')
        res = si_app.get_whoosh_search_suggestion(q)
        self.write(json.dumps(res))
class SearchSuggestionWithFreqJSONHandler(tornado.web.RequestHandler):
    def get(self):
        q = self.get_argument('query')
        res = si_app.get_whoosh_search_suggestion_term_freq(q)
        self.write(json.dumps(res))
'''
class JSON1Handler(tornado.web.RequestHandler):
    def get(self):
        q = self.get_argument('query')
        res = si_app.get_whoosh_search_suggestion(q)
        self.write(json.dumps(res))

        # self.write(json.dumps(res1))
'''

class hdfs_inotify_get_checkpoint_txid(tornado.web.RequestHandler):
    def get(self):
        # Return Code:
        # -2 : Stop the process and wait for next notice
        # -1 : Start from last trx, as in hadoop inotification definition
        # Integer > 0 : the real trx id and proceed.

        #self.write('890')
        #return
        data_source_name = self.get_argument('data_source_name')
        ds_dict = si_app.get_data_source_dict(ds_name=data_source_name)
        if ds_dict['ds_param']['start_inotify'] == 'on' and ds_dict['last_reflect_date'] is not None:
            print('returning:' + str(ds_dict['ds_param']['inotify_trx_id']))
            self.write(str(ds_dict['ds_param']['inotify_trx_id']))
        else: # self.write('-1')
            self.write('-2')

class hdfs_inotify_change(tornado.web.RequestHandler):
    def post(self):
        event_type = self.get_argument('event_type')
        txid = self.get_argument('txid')
        data_source_name = self.get_argument('data_source_name')


        if event_type  == 'CREATE':
            path = self.get_argument('path')
            owner = self.get_argument('owner')
            date_time = self.get_argument('date_time')

            # /data/songs.csv._COPYING_
            if path[-9:] == '_COPYING_':
                # this is a intermidiate file, ignord.
                si_app.logger.info('this file is ignored because ._COPYING_ : ' + path)
                return

            print('event_type', event_type, "time", date_time,'tx:', txid, 'path:', path)

            doc_old = si_app.global_whoosh_search_by_id(q_id=path)
            if len(doc_old) > 0:
                si_app.logger.error('the doc/entity already not exist for creation event.')
                for doc1 in doc_old:
                    si_app.delete_doc_from_index_by_docnum(p_docnum=doc1['docnum'])

            si_app.add_table_content_index(ds_name = data_source_name,
                                           table_id=path,
                                           table_info=json.dumps({'path':path, 'date_time':date_time}),
                                           )


        # Do the thing
        elif event_type == 'UNLINK':
            path = self.get_argument('path')
            date_time = self.get_argument('date_time')

            doc_old = si_app.global_whoosh_search_by_id(q_id=path)
            if len(doc_old) < 1:
                si_app.logger.error('the doc/entity to delete/unlink does not exist.')
            else:
                si_app.delete_doc_from_index_by_docnum(p_docnum=doc_old[0]['docnum'])

            # si_app.commit_index()

            print('event_type', event_type, "time", date_time, 'tx:', txid)
        if event_type in 'RENAME':
            src_path = self.get_argument('src_path')
            dst_path = self.get_argument('dst_path')
            date_time = self.get_argument('date_time')

            doc_old = si_app.global_whoosh_search_by_id(q_id=src_path)
            if len(doc_old) < 1:
                si_app.logger.error('the doc/entity to rename does not exist.' + src_path)
            else:
                si_app.delete_doc_from_index_by_docnum(p_docnum=doc_old[0]['docnum'])

            doc_new = si_app.global_whoosh_search_by_id(q_id=dst_path)
            if len(doc_new) > 0:
                si_app.logger.error('the doc/entity already not exist for creation event.')
                for doc1 in doc_new:
                    si_app.delete_doc_from_index_by_docnum(p_docnum=doc1['docnum'])


            si_app.add_table_content_index(ds_name = data_source_name,
                                           table_id=dst_path,
                                           table_info=json.dumps({'path':dst_path, 'date_time':date_time}),
                                           )
            # si_app.commit_index()
            path = src_path + '.....' + dst_path
            print('event_type', event_type,'path',path, "time", date_time,'tx:', txid)
        else:
            print('event_type', event_type )

        print(path)

        # si_app.data_source_dict[data_source_name]['ds_param']['inotify_trx_id'] = txid


        #there may be performance issues if we update trxid everytime to database. Therefore, i updated to cache
        # most of times.

        # si_app.update_data_soruce(ds_dict=si_app.data_source_dict[data_source_name])
        si_app.update_data_soruce_trx_id(ds_name=data_source_name, trx_id=txid)

        self.write('received!' + path)










#!/usr/bin/env python

import os.path

import tornado.escape
import tornado.httpserver
import tornado.ioloop
import tornado.ioloop
import tornado.options
import tornado.web
import tornado.web

from . import config
from . import handler

# from app import dbmodels

class MApplication(tornado.web.Application):
    def __init__(self):
        handlers = [
            (r"/", handler.OverviewHandler),
            (r"/overview", handler.OverviewHandler),
            (r"/database_json", handler.DatabaseJSONHandler),
            (r"/database_summary", handler.DatabaseSummaryHandler),
            (r"/add_data_source", handler.AddDataSourceHandler),
            (r"/delete_data_source", handler.DeleteDataSourceHandler),
            (r"/relfect_data_source", handler.ReflectDataSourceHandler),
            (r"/view_table_in_notebook", handler.ViewTableInNotebookHandler),
            #

            (r"/reload_plugins", handler.ReloadPluginHandler),
            (r"/global_search", handler.GlobalSearchHandler),
            (r"/search_suggestion_json", handler.SearchSuggestionJSONHandler),
            (r"/search_suggestion_freq", handler.SearchSuggestionWithFreqJSONHandler),

            (r"/hdfs_inotify_change", handler.hdfs_inotify_change),
            (r"/hdfs_inotify_get_checkpoint_txid", handler.hdfs_inotify_get_checkpoint_txid),

        ]
        settings = dict(
            website_title=u"Schema Index",
            template_path=os.path.join(os.path.dirname(__file__), "templates"),
            static_path=os.path.join(os.path.dirname(__file__), "static"),
            # ui_modules={"Entry": EntryModule},
            xsrf_cookies=False,
            cookie_secret="databasemodels:_GENERATE_YOUR_OWN_RANDOM_VALUE_HERE__",
            login_url="/auth/login",
            debug=True,
        )
        super(MApplication, self).__init__(handlers, **settings)
        # Have one global connection to the blog DB across all handlers
        # self.engine = engine

def run_webserver(addr = None, port=8088):
    tornado.options.parse_command_line()
    app = tornado.httpserver.HTTPServer(MApplication())
    if addr is None:
        app.listen(port=port)
    else:
        app.listen(port=port, address=addr)

    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    run_webserver(port = config.cfg['web']['port'], addr = config.cfg['web']['address'])


'''    tornado.options.parse_command_line()
    app = tornado.httpserver.HTTPServer(MApplication())
    app.listen(config.cfg['web']['port'])
    tornado.ioloop.IOLoop.current().start()
'''


#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" stanmoctl.py
    
    Usage:
        schemaindex -h
        schemaindex runserver [--port=<port> ] [--ip=<ip> ] [--browser]
        schemaindex list ( data_source | plugin )
        schemaindex add <data_source_name> --plugin=<spec_name> --ds_param=<ds_param_string>
        schemaindex init
        schemaindex reload plugin
        schemaindex reflect <data_source_name>
        schemaindex snippet <data_source_name> <table_id>
        schemaindex search <search_key_word>...
        schemaindex show <data_source_name>

    Options:
        -h                    : show this help message
        data_source_name      : name of data source given when it was created
        spec_name             : The name of model specification
        --input_file=<path>   : Path of input files to the model
        --output_file=<path>  : Path to write the model output file.
        --ip=<ip>             : The ip address to listen on, and --ip=* indicates all interfaces.
        --port=<port>         : The OS port where server will be listening on. It uses 5000 if omitted..
        --browser             : Open the browser automatically
"""
# This command push will be implemented later.
#        schemaindex push [specs]
#        schemaindex pull [specs]  # To download a plugin from central repository
#        schemaindex search [specs]
# following: https://github.com/docopt/docopt

import os
from docopt import docopt
from schemaindex.app.schemaindexapp import si_app #
from schemaindex.app.webserver import run_webserver
from schemaindex.app.pluginmanager import si_pm

def initialize_schemaindex():
    db_file_path = si_app.config['database']['sqlite_file']
    to_init_indicator = db_file_path  # cfg['main']['init_indicator_file']
    si_app.logger.debug('Checking whether to re-initialize by file:' + to_init_indicator)

    if os.path.exists(to_init_indicator):
        si_app.logger.debug('DB file is ready, no re-init, going to normal startup.')
        return;
    print("\n\r\n\r")
    si_app.logger.debug('SchemaIndex platform will be re-initialized because there no data file yet.')
    print('SchemaIndex platform will be re-initialized because there no data file yet.')

    si_app.schemaindex_init(db_file_path = db_file_path)
    si_pm.scan_reflect_plugins()

    try:
        si_app.init_notebook_extensions()
    except Exception as e:
        print('Exceptions during notebook extension installation: {0}'.format(e.strerror))



    si_app.logger.debug('SchemaIndex platform is initialized.')


# os.remove(to_init_indicator)





def list_plugs():
    plist = si_app.get_plugin_list()

    # {"name":spec_name, "path":os.path.join(model_spec_path,item)} )
    if len(plist) > 0:
        print('{0:40} {1:35} '.format('Reflect Plugin Name:', 'Path:  '))
        for spec in plist:
            print('{0:40} {1:55}'.format(spec.plugin_name, os.path.join(spec.plugin_spec_path, spec.module_name)))
    else:
        print('No plugins are found!')



def main():
    """ main-entry point for schemaindex program, parse the commands and build the si_app platform """
    docopt_args = docopt(__doc__)

    # First we initialize the schemaindex database and load the plugins
    initialize_schemaindex()

    # Parse the User command and the required arguments
    if docopt_args["list"]:
        if docopt_args["data_source"] == True:
            # print(json.dumps(si_app.list_models()))
            si_app.list_data_sources()
        elif docopt_args["plugin"] == True:
            list_plugs()


    # Parse the User command and the required arguments
    # create <model_name> --plugin=<spec_name>
    if docopt_args["add"]:
        model_name = docopt_args["<model_name>"]
        spec_name = docopt_args["--plugin"]
        if spec_name is None:
            print("Please specify the <model_name> and <spec_name>")
        si_app.create_model(model_name=model_name, spec_name=spec_name)
        print("Model is created successfully.")

    elif docopt_args["snippet"]:
        # to reflect the specified data source.
        data_source_name = docopt_args["<data_source_name>"]
        table_id = docopt_args["<table_id>"]
        the_snippet = si_pm.generate_notebook_snippet_for_table(table_id=table_id,
                                                                ds_name=data_source_name)
        print(the_snippet)


    elif docopt_args["reflect"]:
        # to reflect the specified data source.
        data_source_name = docopt_args["<data_source_name>"]
        the_model = si_app.reflect_db(data_source_name=data_source_name)

    elif docopt_args["search"]:
        # to search by a keyword
        q = docopt_args["<search_key_word>"]
        print(q)
        tt = si_app.global_whoosh_search_formatted(q=' '.join(q))
        import json
        print(json.dumps(tt))



    elif docopt_args["runserver"]:
        # to run a HTTP server to provide restful api services..

        if docopt_args["--port"] is not None:
            port =  docopt_args["--port"]
        else:
            port = si_app.config['web']['port']

        addr_ip = 'localhost'
        url = 'http://%s:%s/' % (str(addr_ip), str(port))
        if docopt_args["--ip"] is not None:
            addr_ip =  docopt_args["--ip"]
            url = 'http://%s:%s/' % (str(addr_ip), str(port))
            if docopt_args["--ip"] == '*':
                url = 'http://%s:%s/' % ('localhost', str(port))  # 'http://localhost:' + str(port) + '/'

        si_pm.datasource_init()

        print("\n\rServer started, please visit : " + url)

        if docopt_args["--browser"] == True:
            import webbrowser
            # Open URL in new window, raising the window if possible.
            webbrowser.open_new(url)

        run_webserver(addr=addr_ip, port = port)

    elif docopt_args["reload"]:
        # to reload from plugin folder
        if docopt_args["plugin"] == True:
            si_pm.scan_reflect_plugins()
            print("Plugins are reloaded.")
            list_plugs()

    elif docopt_args["init"]:
        # to reload from plugin folder
        si_app.scan_reflect_plugins()
        if docopt_args["plugin"] == True:

            print("Plugins are reloaded.")
            list_plugs()


    elif docopt_args["show"]:
        # to predict according to trained a model, given the input file.
        port = 5011
        if docopt_args["--port"] is not None:
            port =  docopt_args["--port"]

        model_name = docopt_args["<model_name>"]
        the_model = si_app.load_model(model_name=model_name)
        try:
            the_model.show(port=port)
        except :
            print('Failed to show the model: {0}'.format(e.strerror))


# START OF SCRIPT
if __name__ == "__main__":
    main()






from __future__ import unicode_literals
# according to http://python-future.org/compatible_idioms.html#unicode-text-string-literals
#
#
# #import atexit
import os
import shutil
import logging
import json
import time
import importlib
from builtins import str
# from subprocess import Popen
import subprocess

from sqlalchemy import Column, DateTime, String, Integer, func
from sqlalchemy import create_engine
from sqlalchemy.orm import create_session


from whoosh import index
from whoosh.qparser import QueryParser
from whoosh.fields import *
from whoosh.analysis import FancyAnalyzer

from .config import cfg, SchemaIndexPluginError
from .dbmodels import MColumn, MTable, MDatasource, MPlugin, Base

#sys.path.append( cfg['main']['schemaflex_spec'])

import sqlite3

si_db_engine =  create_engine('sqlite:///' + cfg['database']['sqlite_file'])


class SchemaIndexApp:
    """ The runtime platform for running all mining models
    """

    MODEL_DATAFRAME_DIR = 'data'
    MODEL_INSTANCE_DIR = 'instance'
    MODEL_SPEC_DIR = 'plugin' # Save the specs of a specific model, including file mining_model.json
    MODEL_SPEC_PATH = 'plugin' # Save all model_spec programs. One plugin may serve multiple models
    MODEL_SPEC_FILENAME = 'mining_model.json'
    TIME_FORMATER = "%Y-%m-%d %H:%M:%S"

    config = cfg
    schemaindex_home = cfg['main']['schemaflex_home']

    logger = logging.getLogger('stanmo_logger')

    indexdir = cfg['main']['schemaflex_text_index_path']
    ix = None
    index_writer = None

    data_source_dict = {} # {'__sample__': {}}

    db_session = create_session(bind=si_db_engine)
    data_source_process = {} #  {'__sample__': []}
    time_to_commit = time.time()


    def __init__(self):
        # Add the plugin (model specs) home to sys path for dynamic loading all model specs defined under $STANMO_HOME/plugin
        # sys.path.append(os.path.join(self.schemaindex_home, self.MODEL_SPEC_PATH))

        # self.schemaindex_init()
        #
        # self.datasource_init() # Calling here cause troubles to __import__, weird. 20171222
        pass


    def schemaindex_init(self, db_file_path = None):
        try:
            generate_notebook_dir = self.config['main']['schemaindex_notebooks']
            if not os.path.exists(generate_notebook_dir):
                self.logger.debug('creating folder for generated notebook.')
                os.mkdir(generate_notebook_dir)

            if os.path.exists(db_file_path):
                self.logger.debug('Trying to remove existing db file.')
                os.remove(db_file_path)
            # recreate the sqlite database file
            self.logger.debug('creating db at ...:  ' + db_file_path)
            engine = create_engine('sqlite:///' + db_file_path)
            Base.metadata.create_all(engine)

        except Exception as e:
            print(str(e))
            self.logger.debug('init error: ' + str(e))

        self.logger.debug('trying to re-construct text index at folder: ' + self.indexdir)  # will not print anything
        if os.path.exists(self.indexdir):
            shutil.rmtree(self.indexdir)
        os.mkdir(self.indexdir)

        schema = Schema(ds_name=ID(stored=True),
                        table_id=ID(stored=True),
                        table_info=ID(stored=True),
                        table_content_index=TEXT(stored=False, analyzer=FancyAnalyzer()),
                        ) # , column_info=TEXT(stored=True)
        ix = index.create_in(self.indexdir , schema)
        print("schemaindex platform is re-initialized, with new sqlite data file and whoosh index." )  # will not print anything

    def init_notebook_extensions(self):

        self.logger.debug('SchemaIndex is trying to set up nb and server extensions to jupyter notebook ...')
        nbext_dir = os.path.join( self.schemaindex_home, 'jupyterext','nbext')
        serverext_dir = os.path.join( self.schemaindex_home, 'jupyterext','serverext')

        p = subprocess.call(["jupyter", "nbextension", "install", 'schemaindex_nbext', '--user'],
                   cwd=nbext_dir)
        p = subprocess.call(["jupyter", "nbextension", "enable", 'schemaindex_nbext/main', '--user'],
                   cwd=nbext_dir)

        p = subprocess.call(["jupyter", "serverextension", "enable", '--py', 'schemaindex.jupyterext.serverext.schemaindex_serverext'],
                   cwd=serverext_dir)
        print('SchemaIndex finished installing nb and server extensions to jupyter notebook ...')

    def datasource_init(self):

        #self.logger.debug('SchemaIndex is collecting all data source ...')
        ds_list = self.get_data_source_rs()

        # Here we loop through all data sources. For each of them, if it is not cached, then it is the first time to run schemaindexapp.
        # We do 2 things for each data source
        # 1. cache the ds_dict
        # 2. run datasource_init of each datasource as defined by plugin.
        for row in ds_list:
            if row.ds_name not in self.data_source_dict.keys() :
                self.data_source_dict[row.ds_name] = self.get_data_source_dict(ds_name=row.ds_name)


    def __del__(self):
        if self.index_writer is not None:
            # print('closing index')
            self.index_writer.cancel()
            self.ix.close() #


    def delete_data_soruce(self,ds_dict = None):
        session = create_session(bind=si_db_engine)
        session._model_changes={}

        tab_result = session.query(MTable).filter_by(ds_name=ds_dict['ds_name'])
        if tab_result.count() > 0 and (not ds_dict['delete_reflected_database_automatic'] == 'on'):
            return {'message_type': 'danger',
                    'message_title': 'Error',
                    'message_body':'the data source "' + ds_dict['ds_name'] + '" contains metadata. Please check " Delete Contents in Data Source" and try again'}

        session.begin()
        session.query(MDatasource).filter_by(ds_name=ds_dict['ds_name']).delete(synchronize_session=False)
        session.query(MTable).filter_by(ds_name=ds_dict['ds_name']).delete(synchronize_session=False)
        session.query(MColumn).filter_by(ds_name=ds_dict['ds_name']).delete(synchronize_session=False)
        session.commit()
        self.delete_doc_from_index_by_datasource(ds_name=ds_dict['ds_name'])

        return {'message_type': 'info',
                'message_title': 'Info',
                'message_body': 'the data source "' + ds_dict['ds_name'] + '" is deleted.'}

    def add_data_soruce(self,ds_dict = None):
        session = create_session(bind=si_db_engine)
        session._model_changes={}

        session.begin()
        session.add_all([
            MDatasource(
                               ds_name=ds_dict['ds_name'] ,
                               ds_type=ds_dict['ds_type'] ,
                               ds_param=json.dumps(ds_dict['ds_param']) ,
                               nbr_of_tables=0,
                               nbr_of_columns=0,
                               ds_desc=ds_dict['ds_desc'],
                               created_date = func.now(),
                    )
        ])
        session.commit()
        dbrs1 = session.query(MDatasource).filter_by(ds_name=ds_dict['ds_name'])
        db = dbrs1.first()
        if db is None: # This should not happend!
            si_app.logger.error('error: did not find database')
            return False
        return True

    def update_data_soruce_trx_id(self,ds_name = None, trx_id = None):

        current_time = time.time()
        self.data_source_dict[ds_name]['ds_param']['inotify_trx_id'] = trx_id
        if current_time - self.time_to_commit > self.config['main']['interval_commit_trx_id']:
            self.logger.info('time out, this time write to database')
            self.update_data_soruce(ds_dict=self.data_source_dict[ds_name])
            self.time_to_commit = current_time


    def update_data_soruce(self,ds_dict = None):

        self.data_source_dict[ds_dict['ds_name']] = ds_dict

        session = create_session(bind=si_db_engine)
        session._model_changes={}
        session.begin()
        session.query(MDatasource).filter_by(ds_name=ds_dict['ds_name']).\
            update({
                    'ds_type': ds_dict['ds_type'],
                    'ds_param': json.dumps(ds_dict['ds_param']),
                    'ds_desc':ds_dict['ds_desc']})


        session.commit()
        dbrs1 = session.query(MDatasource).filter_by(ds_name=ds_dict['ds_name'])
        db = dbrs1.first()
        if db is None:
            print('error: did not find database')
        return db


    '''
        def get_data_source_name_list(self):
    
            rs = self.db_session.query(MDatasource) #.filter_by(name='ed')
    
            ds_list = []
            for row in rs:
                ds_list.append({'name': row.ds_name , 'ds_type': row.ds_type } )
            return  ds_list
    '''
    def global_whoosh_search(self, q = ''):

        ix = index.open_dir(self.indexdir)
        res = []
        with ix.searcher() as searcher:
            query = QueryParser("table_content_index", ix.schema).parse(q)
            results = searcher.search(query)
            for r in results:
                res.append({'ds_name': r['ds_name'],
                            'docnum': r.docnum,
                            'table_id':r['table_id'],
                            'table_info': r['table_info'],
                            })
        return res


    def get_schemaindex_statistics(self):

        ix = index.open_dir(self.indexdir)
        res = []
        result_dict = {}
        with ix.searcher() as searcher:

            table_count = searcher.doc_count()
            result_dict['table_count'] = table_count

        ds_resultset = self.get_data_source_rs()
        result_dict['ds_count'] = ds_resultset.count()
        return  result_dict

    def global_whoosh_search_formatted(self, q = ''):
        ix = index.open_dir(self.indexdir)
        res = {}
        with ix.searcher() as searcher:
            query = QueryParser("table_content_index", ix.schema).parse(q)
            results = searcher.search(query, limit=self.config['main']['search_result_limit'])
            for rrr in results:
                ds_dict = si_app.get_data_source_dict(ds_name=rrr['ds_name'])
                if ds_dict['metadata_type'] not in res.keys():
                    res[ds_dict['metadata_type']] = [] # 'table' or 'file'
                res[ds_dict['metadata_type']].append({'ds_name': rrr['ds_name'],
                            'docnum': rrr.docnum,
                            'table_id':rrr['table_id'],
                            'table_info': rrr['table_info'],
                            })
        return res

    def global_whoosh_search_by_id(self, q_id = ''):

        ix = index.open_dir(self.indexdir)
        res = []
        with ix.searcher() as searcher:
            query = QueryParser("table_id", ix.schema).parse(q_id)
            results = searcher.search(query)
            for r in results:
                res.append({'ds_name': r['ds_name'],
                            'docnum': r.docnum,
                            'table_info': r['table_info'],
                            })

        return res

    def delete_doc_from_index_by_docnum(self, p_docnum = None):

        ix = index.open_dir(self.indexdir)
        index_writer = ix.writer()

        index_writer.delete_document(docnum = p_docnum)
        index_writer.commit()
        return True

    def delete_doc_from_index_by_datasource(self, ds_name = None):

        ix = index.open_dir(self.indexdir)
        index_writer = ix.writer()

        ix = index.open_dir(self.indexdir)
        res = []
        with ix.searcher() as searcher:
            query = QueryParser("ds_name", ix.schema).parse(ds_name)
            num = index_writer.delete_by_query(q=query)
            si_app.logger.info( '{0:10} documents from data source {1:15}  '.format( str(num), ds_name) )

        index_writer.commit()
        return True



    def get_whoosh_search_suggestion(self, q = ''):

        indexdir = self.indexdir
        ix = index.open_dir(indexdir)

        result = []
        with ix.reader() as r:
            for aterm in r.most_frequent_terms("table_content_index", number=5, prefix=q):
                one_result = aterm[1].decode('utf-8')
                result.append(one_result) # The result was like (1.0, 'dept_manager'), but here i need only a keyword
                #print(aterm)

        return result

    def get_whoosh_search_suggestion_term_freq(self, q = ''):

        indexdir = self.indexdir
        ix = index.open_dir(indexdir)

        result = []
        with ix.reader() as r:
            for aterm in r.most_frequent_terms("table_content_index", number=5, prefix=q):
                term_name = aterm[1].decode('utf-8')
                term_freq = aterm[0]
                result.append({'table_id':term_name, 'table_freq':term_freq}) # The result was like (1.0, 'dept_manager'), but here i need only a keyword
                #print(aterm)

        return result


    def add_table_content_index(self, table_content_index,ds_name = None, table_id = None, table_info = None):

        ix = index.open_dir(self.indexdir)
        index_writer = ix.writer()
        # index_writer.add_document(ds_name = unicode(ds_name), table_id=unicode(table_id), table_info=unicode(table_info))

        index_writer.add_document(ds_name = str(ds_name)
                                  , table_id=str(table_id)
                                  , table_info=str(table_info)
                                  , table_content_index=str(table_content_index)
                                  )
        index_writer.commit()

    def commit_index(self, table_id=None, table_info=None):
        return
        self.index_writer.commit()
        self.ix = index.open_dir(self.indexdir)
        self.index_writer = self.ix.writer()



    def get_data_source_rs(self):

        rs = self.db_session.query(MDatasource).order_by(MDatasource.ds_name.asc()) #.filter_by(name='ed')
        return rs

    def get_data_source_dict(self, ds_name = None):

        if ds_name in self.data_source_dict.keys():
            return self.data_source_dict[ds_name]

        rs = self.db_session.query(MDatasource).filter_by(ds_name=ds_name)
        if rs.count() > 0:
            ds = rs.first()
            ds_loaded = ds.to_dict() # {x.name: getattr(ds, x.name) for x in ds.__table__.columns}
            mt = self.db_session.query(MDatasource, MPlugin).filter(MDatasource.ds_type == MPlugin.plugin_name).\
                    filter(MDatasource.ds_name==ds_name)
            for da, mp in mt:
                ds_loaded['metadata_type'] = mp.metadata_type
            self.data_source_dict[ds_name] = ds_loaded
            return self.data_source_dict[ds_name]

        return None


    def get_table_list_for_data_source_rs(self, param_ds_name = ''):
        sql = '''
                SELECT c.ds_name,  c.table_name, t.table_comment, group_concat(c.column_name)   as column_names
                FROM mcolumn c, mtable t
                where c.table_name = t.table_name and c.ds_name = t.ds_name and t.ds_name = \'''' + param_ds_name + '''\'
                GROUP BY c.ds_name, c.table_name;
                '''

        tabrs = si_db_engine.execute(sql)
        return  tabrs




    def list_data_sources(self):
        model_list = []
        models = []

        session = create_session(bind=si_db_engine)
        dbrs = session.query(MDatasource)  # .filter_by(name='ed')


        for db in dbrs:
                models.append(db)
        if len(models) > 0:
            print('{0:20}   {1:20}  {2:35}  '.format('data source name',
                                                                           'data source type',
                                                                           'URL'
                                                                           ))
            for a_model in models:
                print('{0:20}   {db_type_name:20}  {table_group1_name:35}  '.format(a_model.ds_name,
                                                                          db_type_name=a_model.ds_type,
                                                                          table_group1_name = a_model.ds_param
                                                                                           )
                      )
        else:
            print('No data source is found!')

        logging.getLogger('stanmo_logger').debug('discovered models: ' + model_list.__str__())
        return model_list


    def get_plugin_name_list(self):
        plugins = []
        rs = self.db_session.query(MPlugin)
        for p in rs:
            plugins.append(p.plugin_name)
        return  plugins

    def get_plugin_list(self):
        plugins = []
        rs = self.db_session.query(MPlugin)
        for p in rs:
            plugins.append(p)
        return  plugins

    def get_plugin_info(self, p_plugin_name = ''):

        p = self.db_session.query(MPlugin).filter_by(plugin_name=p_plugin_name).first()
        # p['ds_param'] = json.loads(p.ds_param )
        return  p


si_app = SchemaIndexApp()



from setuptools import setup, find_packages


# following https://python-packaging-user-guide.readthedocs.io/tutorials/distributing-packages/

setup(name='schemaindex',
      version='0.243',
      description='An indexing engine for different types of data sources, including HDFS, Mysql, etc.',
      url='https://github.com/qiyangduan/schemaindex',
      author='Duan Qiyang',
      author_email='qiyang.duan@gmail.com',
      license='Apache Software License',
      include_package_data=True,
      packages=['schemaindex'
               # ,'schemaindex.plugin'
               ,'schemaindex.jupyterext'
               ,'schemaindex.app'
                ],
      install_requires=[
           'tornado',
           'docopt',
           'sqlalchemy',
           'whoosh',
           'hdfs',
           'future',
          'pymssql',
          'pymysql',



      ],
      entry_points={
          'console_scripts': ['schemaindex=schemaindex.schemaindexctl:main'],
      },
      classifiers = [
          'Development Status :: 3 - Alpha',

          # Indicate who your project is intended for
          'Intended Audience :: Developers',
          'Topic :: System :: Systems Administration',
          'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.6',
      ],
      zip_safe=False)



import os
import json
import subprocess
import datetime

from hdfs import Client

from schemaindex.app.schemaindexapp import si_app

# -----------------------------------
plugin_name = 'simplehdfsindex'
plugin_desc = '''Given the connection information and initial path, this plugin recursively scan through all files in a HDFS system and index them. '''
required_packages = ['hdfs']

ds_param = {'hdfs_web_url': {'type':'String',
                             'full_name':'HDFS Web URL (Required):',
                             'default_value': 'http://localhost:50070',
                             'desc':'''The HDFS web URL, usually staring with HTTP, for example: http://localhost:50070 .  This URL is used by hdfscli library to connect to HDFS web and scan HDFS structure.''',
                             },
            'root_path': {'type':'String',
                             'full_name':'Root path (Required):',
                             'default_value': '/',
                             'desc':'The starting point to index for the index, for example, you can use / or /user', #
                             },

            }


_________ds_param_desc = {'hdfs_web_url': '''The HDFS web URL, usually staring with HTTP, for example: http://localhost:50070 .  This URL is used by hdfscli library to connect to HDFS web and scan HDFS structure.''',
            'root_path': 'The starting point to index for the index', # Sample: /
            }


supported_ds_types = ['HDFS']
metadata_type='file'

sample_ds_url = 'http://localhost:50070'
author = 'duan'
notebook_template_path = 'show_hdfs_file_template.ipynb'



# -----------------------------------
class ReflectEngine():
    ds_dict = None
    si_app = None

    def __init__(self,si_app = None, ds_dict=None, ds_name = None):
        self.si_app = si_app
        if ds_dict is not None:
            self.ds_dict = ds_dict
        elif ds_name is not None:
                self.ds_dict = si_app.get_data_source_dict(ds_name = ds_name)
        else:
            si_app.logger.error('no data source is given!')
            return

    # get the all file information from HDFS
    def QueryHDFSFile(self, pDirectory, pClient, filelist):

        to_list_dir = pDirectory if len(pDirectory) > 1 else '/'
        tDirectoryStatus = pClient.status(to_list_dir)
        tDirectoryType = tDirectoryStatus.get('type')

        if tDirectoryType == 'FILE':
            formatted_time = datetime.datetime.fromtimestamp(int(tDirectoryStatus.get('modificationTime')) / 1000
                                                             ).strftime(si_app.TIME_FORMATER)
            fileDict = {
                'table_name':pDirectory,
                'modificationTime':   formatted_time   ,
                'length': tDirectoryStatus.get('length'),
            }
            filelist.append(fileDict)

        elif tDirectoryType == 'DIRECTORY' and (not pClient.list(to_list_dir)):  # This is an empty folder
            pass
            # I skip empty directories.
        else:

            tDirectorys = pClient.list(to_list_dir)

            #fileNumInFolder = 0  # the number of files in this folder
            tSubDirectorys = []
            for tDirectory in tDirectorys:
                tSubDirectory = pDirectory + '/' + tDirectory
                tSubDirectorys.append(tSubDirectory)
                #if pClient.status(tSubDirectory).get('type') == 'FILE':
                #    fileNumInFolder += 1

            for tSubDirectory in tSubDirectorys:
                self.QueryHDFSFile(tSubDirectory, pClient, filelist)

    # get the all file information from HDFS
    def getHDFSFileInfo(self,tclient, pInitDir):
        filelist = []
        # if pInitDir is a file, fileNum = 1; if pInitDir is a directory, fileNum here has no influence
        self.QueryHDFSFile(pInitDir, tclient, filelist)
        return filelist

    def reflect(self, reload_flag = False):
        if not self.ds_dict:
            si_app.logger.error('error: ds_dict must be provided.')
            return

        if reload_flag:
            si_app.delete_doc_from_index_by_datasource(ds_name=self.ds_dict['ds_name'])


        tclient = Client( self.ds_dict['ds_param']['hdfs_web_url'])     # self.ds_dict['ds_url'])
        path_hdfs = self.ds_dict['ds_param']['root_path']      # self.ds_dict['table_group_name']

        if path_hdfs[-1] == '/':
            # remove trailing '/' for concatenating more path
            path_hdfs = path_hdfs[:-1] # remove the trailing '/'

        filelist = self.getHDFSFileInfo(tclient, path_hdfs)


        for fd in filelist:
            si_app.add_table_content_index(ds_name = self.ds_dict['ds_name'],
                                           table_id=fd['table_name'],
                                           table_info=(json.dumps(fd) ),
                                           table_content_index = ' '.join([fd[k] for k in fd.keys() ])
                                           )
        # ds_dict = si_app.get_data_source_dict(ds_name=self.ds_dict['ds_name'])

    def datasource_init(self):
        pass


    def generate_notebook(self, schemaindex_notebooks_dir = '/tmp', table_id = None):

        spec_dir = si_app.config['main']['schemaflex_spec']
        replace_dict = {'$$file_path$$': table_id,
                        '$$hdfs_web_url$$': self.ds_dict['ds_param']['hdfs_web_url']}
        generated_loc = os.path.join(spec_dir,'hdfsindex', 'show_hdfs_file_%s_%s.ipynb' %
                                     (self.ds_dict['ds_name'], table_id.replace('/','_') )          )
        with open(os.path.join(spec_dir,'hdfsindex', 'show_hdfs_file_template.ipynb'), "rt") as fin:
            with open( generated_loc, "wt") as fout:
                for line in fin:
                    snippet_result = line
                    for key in replace_dict.keys():
                        snippet_result = snippet_result.replace(key, replace_dict[key])

                    fout.write(snippet_result)
        return generated_loc

    def generate_notebook_snippet(self, table_id = None):

        snippet_template = '''
#!pip install pandas        
import pandas as pd
from hdfs import Client
tclient = Client( '$$hdfs_web_url$$')
with tclient.read('$$file_path$$', encoding='utf-8') as reader:
  df=pd.read_csv(reader)
df.head()
'''
        replace_dict = {'$$file_path$$': table_id,
                        '$$hdfs_web_url$$': self.ds_dict['ds_param']['hdfs_web_url']}
        # snippet_result = (reduce(lambda a, kv: a.replace(*kv), replace_dict.iteritems(), snippet_template))
        snippet_result = snippet_template
        for key in replace_dict.keys():
            snippet_result = snippet_result.replace(key, replace_dict[key])

        # print(snippet_result)
        return snippet_result

if __name__ == "__main__":
    ds_dict = {
        'ds_name': 'hdfs2',
        'ds_type': 'simplehdfsindex',
        'ds_desc': 'created by unittest of hdfsindex',
        'ds_url': 'http://localhost:50070',
        'ds_param': {'hdfs_web_url': 'http://localhost:50070' ,
                     'root_path': '/',
                     }
    }

    si_app.add_data_soruce(ds_dict)
    a = ReflectEngine(ds_dict = ds_dict)
    a.reflect()






from __future__ import unicode_literals
# according to http://python-future.org/compatible_idioms.html#unicode-text-string-literals
#
#
# #import atexit
import os
import shutil
import logging
import json
import time
import importlib
from builtins import str


from sqlalchemy import Column, DateTime, String, Integer, func
from sqlalchemy import create_engine
from sqlalchemy.orm import create_session


from whoosh import index
from whoosh.qparser import QueryParser
from whoosh.fields import *


from .config import cfg, SchemaIndexPluginError
from .dbmodels import MColumn, MTable, MDatasource, MPlugin, Base
from .schemaindexapp import si_app




class PluginManager:
    """ One utility class with functions to deal with all the plugins
    """

    MODEL_DATAFRAME_DIR = 'data'
    MODEL_INSTANCE_DIR = 'instance'
    MODEL_SPEC_DIR = 'plugin' # Save the specs of a specific model, including file mining_model.json
    MODEL_SPEC_PATH = 'plugin' # Save all model_spec programs. One plugin may serve multiple models
    MODEL_SPEC_FILENAME = 'mining_model.json'
    TIME_FORMATER = "%Y-%m-%d %H:%M:%S"

    config = cfg
    schemaindex_home = cfg['main']['schemaflex_home']
    logger = logging.getLogger('stanmo_logger')

    si_db_engine = create_engine('sqlite:///' + cfg['database']['sqlite_file'])
    db_session = create_session(bind=si_db_engine)

    def __init__(self):
        # Add the plugin (model specs) home to sys path for dynamic loading all model specs defined under $STANMO_HOME/plugin
        # sys.path.append(os.path.join(self.schemaindex_home, self.MODEL_SPEC_PATH))

        # self.pluginmanager_init()
        #
        #self.datasource_init() # Calling here cause troubles to __import__, weird. 20171222
        self.si_app = si_app
        self.logger.debug('SchemaIndex plugin manager is started.')


    def pluginmanager_initial_loading(self):

            self.logger.debug('scanning plugins for re-initialization... ')
            self.scan_reflect_plugins()
            self.logger.debug('init error: ' + str(e))  # will not print anything



    def datasource_init(self):

        self.logger.debug('SchemaIndex is trying to boot up all init service by each data source ...')
        ds_list = si_app.get_data_source_rs()

        # Here we loop through all data sources. For each of them, if it is not cached, then it is the first time to run schemaindexapp.
        # We do 2 things for each data source
        # 1. cache the ds_dict
        # 2. run datasource_init of each datasource as defined by plugin.
        for row in ds_list:
            if row.ds_name not in si_app.data_source_dict.keys() :
                si_app.data_source_dict[row.ds_name] = si_app.get_data_source_dict(ds_name=row.ds_name)

                the_engine = self.get_reflect_plugin(p_plugin_name=row.ds_type)['reflect_engine']
                one_ds = the_engine.ReflectEngine(ds_dict = row.to_dict())
                one_ds.datasource_init()

    def get_reflect_plugin(self, p_plugin_name = None):
        p = self.db_session.query(MPlugin).filter_by(plugin_name=p_plugin_name).first()
        return self.load_reflect_engine(p.module_name)

    '''
        def list_reflect_plugins(self):
            logger = si_app.logger # logging.getLogger('stanmo_logger')
    
            return spec_list
    '''

    def scan_reflect_plugins(self):

        self.logger.debug('scanning plugins for re-initialization... ')

        plugin_spec_path = os.path.join(self.schemaindex_home, self.MODEL_SPEC_PATH)

        self.logger.debug('looking for model plugin in path: ' + plugin_spec_path)
        spec_list = []

        for item in os.listdir(plugin_spec_path):
            if os.path.isdir(os.path.join(plugin_spec_path, item)):
                # to avoid : schemaindex.app.config.SchemaIndexPluginError: module '__pycache__' has no attribute 'plugin_name'
                if (item == '__pycache__'):
                    continue
                a_plugin = self.load_reflect_engine(item, plugin_spec_path=plugin_spec_path)
                spec_list.append(a_plugin)


        self.db_session.begin()
        self.db_session.query(MPlugin).delete()
        for plugin_dict in spec_list:
            if plugin_dict is not None:
                self.db_session.add_all([
                    MPlugin( plugin_name=plugin_dict['plugin_name'] ,
                                      module_name=plugin_dict['module_name'] ,
                                      plugin_spec_path=plugin_dict['plugin_spec_path'],
                                      metadata_type = plugin_dict['metadata_type'],
                                      ds_param=json.dumps(plugin_dict['ds_param']),
                                      supported_ds_types=plugin_dict['supported_ds_types'],
                                      notebook_template_path=plugin_dict['notebook_template_path'],
                                      # sample_ds_url=plugin_dict['sample_ds_url'],
                                      author=plugin_dict['author'],
                                      plugin_desc=plugin_dict['plugin_desc'],
                                    )
                                    ])
        self.db_session.commit()



    def load_reflect_engine(self, dottedpath, plugin_spec_path = None):

        assert dottedpath is not None, "dottedpath must not be None"

        try:
            module = __import__(dottedpath, globals(), locals(), [])
            return {'reflect_engine': module,
                    'module_name': dottedpath,
                    'plugin_name': getattr(module, 'plugin_name'),
                    'metadata_type': getattr(module, 'metadata_type'),
                    'ds_param': getattr(module, 'ds_param'),
                    'plugin_spec_path': plugin_spec_path,
                    'notebook_template_path': getattr(module, 'notebook_template_path'),
                    # 'sample_ds_url': getattr(module, 'sample_ds_url'),
                    'plugin_desc': getattr(module, 'plugin_desc'),
                    'supported_ds_types': json.dumps(getattr(module, 'supported_ds_types')) ,
                    'author': getattr(module, 'author'),
                    }
        except Exception as e:
            self.logger.error( "load_reflect_engine (error): Failed to import plugin %s, due to error :{%s}" % (dottedpath, e) )
            # return None
            raise SchemaIndexPluginError(e)



    def reflect_db(self,data_source_name=None):
        session = self.db_session
        dbrs = session.query(MDatasource).filter_by(ds_name=data_source_name)
        for row in dbrs:
            the_engine= self.get_reflect_plugin(row.ds_type)['reflect_engine']
            a_ds = the_engine.ReflectEngine(ds_dict = si_app.get_data_source_dict(ds_name= row.ds_name) ) #SQLAlchemyReflectEngine()
            a_ds.reflect(reload_flag=True)

            session_update = create_session(bind=self.si_db_engine)
            session_update.begin()
            session_update.query(MDatasource).filter_by(ds_name=row.ds_name).\
                update({'last_reflect_date': func.now()}, synchronize_session='fetch')
            session_update.commit()
            session_update.close()
        session.close()

    def generate_notebook_for_table(self, table_id = None, ds_name = None):

        ds_dict = si_app.get_data_source_dict(ds_name= ds_name)
        the_engine = self.get_reflect_plugin(ds_dict['ds_type'])['reflect_engine']

        a_ds = the_engine.ReflectEngine(ds_dict = ds_dict )
        generate_notebook_loc = a_ds.generate_notebook(schemaindex_notebooks_dir=self.config['main']['schemaindex_notebooks']
                               , table_id= table_id)
        return  generate_notebook_loc

    def generate_notebook_snippet_for_table(self, table_id = None, ds_name = None):

        ds_dict = si_app.get_data_source_dict(ds_name= ds_name)
        if ds_dict is None:
            return None
        the_engine = self.get_reflect_plugin(ds_dict['ds_type'])['reflect_engine']
        the_ds = the_engine.ReflectEngine(ds_dict = ds_dict )
        generate_notebook_snippet = the_ds.generate_notebook_snippet(table_id= table_id)
        return  generate_notebook_snippet


si_pm = PluginManager()



from sqlalchemy import create_engine
from sqlalchemy.orm import create_session
from sqlalchemy import MetaData
from sqlalchemy import Table, Column, DateTime, String, Integer, ForeignKey, func
import json
import os

from schemaindex.app.dbmodels import MTable, MColumn, MDatasource
from schemaindex.app.schemaindexapp import si_app


engine = create_engine('sqlite:///' + si_app.config['database']['sqlite_file']  )

class SQLAlchemyReflectEngine():
    ds_dict = None
    def __init__(self,ds_dict=None, ds_name = None):
        if ds_dict is not None:
            self.ds_dict = ds_dict
        elif ds_name is not None:
                self.ds_dict = si_app.get_data_source_dict(ds_name = ds_name)
        else:
            si_app.logger.error('no data source is given!')

    @staticmethod
    def echo(self):
        return 'echo'

    def datasource_init(self):
        pass

    def reflect(self, reload_flag = False):
        if not self.ds_dict:
            print('error: ds_url must be provided.')
            return

        reflect_engine = create_engine(self.ds_dict['ds_param']['connect_string']   )
        metadata = MetaData(bind=reflect_engine)
        metadata.reflect()



        session = create_session(bind=engine)
        session._model_changes={}
        session.begin()

        if reload_flag:
            session.query(MTable).filter_by(ds_name = self.ds_dict['ds_name']).delete()
            session.query(MColumn).filter_by(ds_name = self.ds_dict['ds_name']).delete()
            si_app.delete_doc_from_index_by_datasource(ds_name=self.ds_dict['ds_name'])
        # tables = metadata.tables.values()
        # print(tables)



        for t in metadata.sorted_tables:
            #print(t.name)
            # t.columns.id
            # t.columns.id.comment
            session.add_all([
                MTable(
                        ds_name = self.ds_dict['ds_name'],
                        table_group_name = self.ds_dict['ds_param']['schema_name'],
                        table_name=t.name,
                        table_comment = ''
                        )
            ])


            table1 = Table(t.name, metadata
                           , autoload=True, autoload_with=engine)

            # print([c.name for c in table1.columns])
            #print([c.type for c in table1.columns])
            #print([c for c in table1.columns])
            column_list = []
            column_name_str = ''
            for c in table1.columns:
                # print(c)
                session.add_all([
                    MColumn(
                            ds_name = self.ds_dict['ds_name'],
                            table_name=t.name,
                            table_group_name=self.ds_dict['ds_param']['schema_name'],
                            column_name = c.name,
                            column_type = str(c.type),
                            column_comment = None # c.doc
                )
                ])
                column_list.append([c.name, str(c.type), c.doc])
                column_name_str = column_name_str + ', ' + c.name

            si_app.add_table_content_index(ds_name = self.ds_dict['ds_name'],
                                           table_id=t.name, #'/'.join(['/', self.ds_dict['ds_name'], t.name]),
                                           table_info=(json.dumps({ #"table_group_name":  self.ds_dict['ds_param']['schema_name'],
                                                                     # "ds_name":    self.ds_dict['ds_name'] ,
                                                                     "table_name":  t.name ,
                                                                     # "table_comment":    ' ' ,
                                                                     "column_info": column_list,
                                                                     }
                                                                    )
                                                   ),
                                           table_content_index= t.name + '  ' + column_name_str
                                           )


        session.commit()
        si_app.commit_index()

    def generate_notebook(self, schemaindex_notebooks_dir = '/tmp', table_id = None):

        spec_dir = si_app.config['main']['schemaflex_spec']
        replace_dict = {'$$TABLE$$': table_id,
                        '$$connect_string$$': self.ds_dict['ds_param']['connect_string']}
        generated_loc = os.path.join(spec_dir,'hdfsindex', 'show_table_%s_%s.ipynb' %
                                     (self.ds_dict['ds_name'], table_id.replace('/','_') )          )
        with open(os.path.join(spec_dir,'sqlalchemyindex', 'show_sqlalchemy_table_template.ipynb'), "rt") as fin:
            with open( generated_loc, "wt") as fout:
                for line in fin:
                    # fout.write(line.replace('$$TABLE$$', table_name))
                    snippet_result = line
                    for key in replace_dict.keys():
                        snippet_result = snippet_result.replace(key, replace_dict[key])

                    # fout.write(reduce(lambda a, kv: a.replace(*kv), replace_dict.iteritems(), line))
                    fout.write(snippet_result)

        return generated_loc

    def generate_notebook_snippet(self, table_id = None):

        snippet_template = '''
#!pip install pandas        
from sqlalchemy import create_engine
import pandas as pd

ds_url = '$$connect_string$$'
engine = create_engine(ds_url)
conn = engine.connect()
# for quick glance, append Oracle:  ' where rownum < 3' , Mysql/sqlite: ' limit 10'
df = pd.read_sql('select * from $$TABLE$$', con=conn)
df.head()
'''
        replace_dict = {'$$TABLE$$': table_id,
                        '$$connect_string$$': self.ds_dict['ds_param']['connect_string']}
        # snippet_result = (reduce(lambda a, kv: a.replace(*kv), replace_dict.iteritems(), snippet_template))
        snippet_result = snippet_template
        for key in replace_dict.keys():
            snippet_result = snippet_result.replace(key, replace_dict[key])

        # print(snippet_result)
        return snippet_result

if __name__ == "__main__":
    ds_dict = {
        'ds_name': 'emp1',
        'ds_type': 'hdfsindex',
        'ds_desc': 'created by unittest of hdfsindex',
        'ds_param': {'connect_string': 'mysql://root:xxx@localhost/employees',
                     'schema_name': 'na',
                     }
    }


    adb = SQLAlchemyReflectEngine(ds_dict)
    adb.reflect()





plugin_name = 'oracle'
plugin_desc = 'oracle reflect engine based on sqlalchemyindex, which works for many mainstream databases, including Mysql, Sqlite, etc. For mysql to work, you need to install pymysql/mysqlclient (python3) or mysql-python (python2) in advance.'
supported_ds_types = ['oracle']
required_packages = ['cx_Oracle']



ds_param = {'schema_name': {'type':'String',
                             'full_name':'Schema Name to Reflect:',
                            'default_value':'hr',
                             'desc':'',
                             },
            'connect_string': {'type':'String',
                             'full_name':'Database Connection String:',
                               'default_value': 'oracle+cx_oracle://hr:hr@192.168.1.1:1521/orcl',
                               'desc':'''The sqlalchemy access URL,  for example: oracle://root:xyz@localhost/employees.''',
                             },
            }

sample_ds_url = 'mysql://root:xyz@localhost/employees'
metadata_type='table'

author = 'duan'
class_name = 'sqlalchemyindex.SQLAlchemyReflectEngine'
notebook_template_path = 'show_sqlalchemy_table_template.ipynb'

from schemaindex.plugin.sqlalchemyindex import ReflectEngine as SqlAlchemyReflectEngine
# -----------------------------------
class ReflectEngine(SqlAlchemyReflectEngine):
    pass
#    def __init__(self,ds_dict=None, ds_name = None):
#        SqlAlchemyReflectEngine.__init__(ds_dict = ds_dict, ds_name = ds_name )



from notebook.utils import url_path_join
from notebook.base.handlers import IPythonHandler

import json
import tornado.web

from schemaindex.app.schemaindexapp import si_app  #
from schemaindex.app.pluginmanager import si_pm  #


def _jupyter_server_extension_paths():
    return [{
        "module": "schemaindex.jupyterext.serverext.schemaindex_serverext"
    }]


class SchemaindexGenerateSnippetHandler(IPythonHandler):
    def get(self):
        table_id = self.get_argument('table_id')
        ds_name = self.get_argument('ds_name')
        # print('i got parameter' + table_id + 'i got parameter' + ds_name)

        res = si_pm.generate_notebook_snippet_for_table(table_id=table_id,
                                                        ds_name=ds_name)
        # print('i got result:' + str(res))

        self.write(str(res))  #


class SchemaindexSearchHandler(IPythonHandler):
    def get(self):
        q = self.get_argument('q')
        # print('i got parameter' + q)

        res = si_app.global_whoosh_search(q)
        # print('i got result:' + str(res))

        self.write(json.dumps(res))  # res) #

class SchemaindexSearchFormattedHandler(IPythonHandler):
    def get(self):
        q = self.get_argument('q')
        # print('i got parameter' + q)

        res = si_app.global_whoosh_search_formatted(q)
        # print('i got result:' + str(res))

        self.write(json.dumps(res))  # res) #

class SchemaindexStatisticsHandler(IPythonHandler):
    def get(self):
        res = si_app.get_schemaindex_statistics()
        # print('i got result:' + str(res))
        self.write(json.dumps(res))  # res) #



class SchemaindexSearchSuggestionJSONHandler(tornado.web.RequestHandler):
    def get(self):
        q = self.get_argument('query')
        res = si_app.get_whoosh_search_suggestion_term_freq(q)
        self.write(json.dumps(res))

class SchemaindexBaseURLHandler(tornado.web.RequestHandler):
    def get(self):
        u = str(si_app.nb_web_app.settings)
        self.write( u)

def load_jupyter_server_extension(nb_server_app):
    """
    Called when the extension is loaded.

    Args:
        nb_server_app (NotebookWebApplication): handle to the Notebook webserver instance.
    """
    nb_server_app.log.info("schemaindex server extension is loaded!")
    web_app = nb_server_app.web_app

    si_app.nb_web_app =  web_app # .settings['base_url']
    host_pattern = '.*$'
    route_pattern = url_path_join(web_app.settings['base_url'], '/schemaindex', '/global_search')
    route_handler = [(route_pattern, SchemaindexSearchHandler),
                     (url_path_join(web_app.settings['base_url'], '/schemaindex', '/global_search_formatted'),
                      SchemaindexSearchFormattedHandler),
                     (url_path_join(web_app.settings['base_url'], '/schemaindex', '/search_suggestion_json'),
                      SchemaindexSearchSuggestionJSONHandler),
                     (url_path_join(web_app.settings['base_url'], '/schemaindex', '/generate_snippet'),
                      SchemaindexGenerateSnippetHandler),
                     (url_path_join(web_app.settings['base_url'], '/schemaindex', '/get_schemaindex_base_rul'),
                      SchemaindexBaseURLHandler),
                     (url_path_join(web_app.settings['base_url'], '/schemaindex', '/get_schemaindex_statistics'),
                      SchemaindexStatisticsHandler),
                     ]
    web_app.add_handlers(host_pattern, route_handler)




class GlobalSearchHandler(IPythonHandler):
    def get(self):
        q = self.get_argument('q')
        res = ''
        if q == '1':
            res = '[{"column_info": [["ds_name", "VARCHAR(100)", null], ["table_group_name", "VARCHAR(200)", null], ["table_name", "VARCHAR(255)", null], ["table_type", "VARCHAR(255)", null], ["table_comment", "VARCHAR(5000)", null]], "table_group_name": "na", "table_name": "mtable", "ds_name": "q"}]'
        elif q == '2':
            res = ''' [ {"column_info": [["ds_name", "VARCHAR(100)", null], ["table_group_name", "VARCHAR(200)", null], ["table_name", "VARCHAR(255)", null], ["table_type", "VARCHAR(255)", null], ["table_comment", "VARCHAR(5000)", null]], "table_group_name": "na", "table_name": "mtable", "ds_name": "q"}
                  , 	{"column_info": [["ds_name", "VARCHAR(100)", null], ["nbr_of_tables", "INTEGER", null], ["nbr_of_columns", "INTEGER", null], ["ds_type", "VARCHAR(100)", null], ["created_date", "DATETIME", null], ["last_reflect_date", "DATETIME", null], ["ds_param", "VARCHAR(4000)", null], ["ds_desc", "VARCHAR(1000)", null], ["ds_tags", "VARCHAR(2550)", null]], "table_group_name": "na", "table_name": "mdatasource", "ds_name": "q"}
                  ]
                  '''
        else:
            res = '[]'
        self.write(res)  # json.dumps(res))

    def post(self):
        q = self.get_argument('q')
        print('i got parameter' + q)
        res = ''
        if q == '1':
            res = '[{"column_info": [["ds_name", "VARCHAR(100)", null], ["table_group_name", "VARCHAR(200)", null], ["table_name", "VARCHAR(255)", null], ["table_type", "VARCHAR(255)", null], ["table_comment", "VARCHAR(5000)", null]], "table_group_name": "na", "table_name": "mtable", "ds_name": "q"}]'
        elif q == '2':
            res = ''' [ {"column_info": [["ds_name", "VARCHAR(100)", null], ["table_group_name", "VARCHAR(200)", null], ["table_name", "VARCHAR(255)", null], ["table_type", "VARCHAR(255)", null], ["table_comment", "VARCHAR(5000)", null]], "table_group_name": "na", "table_name": "mtable", "ds_name": "q"}
                  , 	{"column_info": [["ds_name", "VARCHAR(100)", null], ["nbr_of_tables", "INTEGER", null], ["nbr_of_columns", "INTEGER", null], ["ds_type", "VARCHAR(100)", null], ["created_date", "DATETIME", null], ["last_reflect_date", "DATETIME", null], ["ds_param", "VARCHAR(4000)", null], ["ds_desc", "VARCHAR(1000)", null], ["ds_tags", "VARCHAR(2550)", null]], "table_group_name": "na", "table_name": "mdatasource", "ds_name": "q"}
                  ]
                  '''
        else:
            res = '[]'
        self.write(res)  # json.dumps(res))

class SearchSuggestionJSONHandler(tornado.web.RequestHandler):
    def get(self):
        q = self.get_argument('query')
        res = {}
        if q == 'ta':
            res = [{"ds_name": "1", "table_id": "table"},
                   {"ds_name": "2", "table_id": "tags"},
                   ]

        elif q == 'mta':
            res = [{"data": [{"ds_name": "3", "display": "mtable"},
                             ]
                    }]
        else:
            res = [{"ds_name": "5", "table_id": "table"},
                   {"ds_name": "6", "table_id": "tags"},
                   ]

        self.write(json.dumps(res))

    def post(self):
        q = self.get_argument('query')
        res = {}
        if q == 'ta':
            res = {"data": [{"ds_name": "1", "display": "table"},
                            {"ds_name": "2", "display": "tags"},
                            ]
                   }
        elif q == 'mta':
            res = {"data": [{"ds_name": "3", "display": "mtable"},
                            ]
                   }
        else:
            res = {"data": [{"ds_name": "5", "display": "table"},
                            {"ds_name": "6", "display": "tags"},
                            ]
                   }

        self.write(json.dumps(res))
